"""Entities models for UiPath Platform API interactions."""

from enum import Enum
from types import EllipsisType
from typing import Any, Dict, List, Optional, Type, Union, get_args, get_origin

from pydantic import BaseModel, ConfigDict, Field, create_model


class ReferenceType(Enum):
    """Enum representing types of references between entities."""

    ManyToOne = "ManyToOne"


class FieldDisplayType(Enum):
    """Enum representing display types of fields in entities."""

    Basic = "Basic"
    Relationship = "Relationship"
    File = "File"
    ChoiceSetSingle = "ChoiceSetSingle"
    ChoiceSetMultiple = "ChoiceSetMultiple"
    AutoNumber = "AutoNumber"


class DataDirectionType(Enum):
    """Enum representing data direction types for fields in entities."""

    ReadOnly = "ReadOnly"
    ReadAndWrite = "ReadAndWrite"


class JoinType(Enum):
    """Enum representing types of joins between entities."""

    LeftJoin = "LeftJoin"


class EntityType(Enum):
    """Enum representing types of entities."""

    Entity = "Entity"
    ChoiceSet = "ChoiceSet"
    InternalEntity = "InternalEntity"
    SystemEntity = "SystemEntity"


class EntityFieldMetadata(BaseModel):
    """Model representing metadata for an entity field."""

    model_config = ConfigDict(
        validate_by_name=True,
    )
    type: str
    required: bool
    name: str


class ExternalConnection(BaseModel):
    """Model representing an external connection."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    id: str
    connection_id: str = Field(alias="connectionId")
    element_instance_id: str = Field(alias="elementInstanceId")
    folder_id: str = Field(alias="folderKey")  # named folderKey in TS SDK
    connector_id: str = Field(alias="connectorKey")  # named connectorKey in TS SDK
    connector_name: str = Field(alias="connectorName")
    connection_name: str = Field(alias="connectionName")


class ExternalFieldMapping(BaseModel):
    """Model representing an external field mapping."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    id: str
    external_field_name: str = Field(alias="externalFieldName")
    external_field_display_name: str = Field(alias="externalFieldDisplayName")
    external_object_id: str = Field(alias="externalObjectId")
    external_field_type: str = Field(alias="externalFieldType")
    internal_field_id: str = Field(alias="internalFieldId")
    direction_type: DataDirectionType = Field(alias="directionType")


class FieldDataType(BaseModel):
    """Model representing data type information for a field."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    name: str
    length_limit: Optional[int] = Field(default=None, alias="LengthLimit")
    max_value: Optional[int] = Field(default=None, alias="MaxValue")
    min_value: Optional[int] = Field(default=None, alias="MinValue")
    decimal_precision: Optional[int] = Field(default=None, alias="DecimalPrecision")


class FieldMetadata(BaseModel):
    """Model representing metadata for an entity field."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    id: Optional[str] = Field(default=None, alias="id")
    name: str
    is_primary_key: bool = Field(alias="isPrimaryKey")
    is_foreign_key: bool = Field(alias="isForeignKey")
    is_external_field: bool = Field(alias="isExternalField")
    is_hidden_field: bool = Field(alias="isHiddenField")
    is_unique: bool = Field(alias="isUnique")
    reference_name: Optional[str] = Field(default=None, alias="referenceName")
    reference_entity: Optional["Entity"] = Field(default=None, alias="referenceEntity")
    reference_choiceset: Optional["Entity"] = Field(
        default=None, alias="referenceChoiceset"
    )
    reference_field: Optional["EntityField"] = Field(
        default=None, alias="referenceField"
    )
    reference_type: ReferenceType = Field(alias="referenceType")
    sql_type: "FieldDataType" = Field(alias="sqlType")
    is_required: bool = Field(alias="isRequired")
    display_name: str = Field(alias="displayName")
    description: Optional[str] = Field(default=None, alias="description")
    is_system_field: bool = Field(alias="isSystemField")
    field_display_type: Optional[str] = Field(
        default=None, alias="fieldDisplayType"
    )  # Should be FieldDisplayType enum
    choiceset_id: Optional[str] = Field(default=None, alias="choicesetId")
    default_value: Optional[str] = Field(default=None, alias="defaultValue")
    is_attachment: bool = Field(alias="isAttachment")
    is_rbac_enabled: bool = Field(alias="isRbacEnabled")


class ExternalField(BaseModel):
    """Model representing an external field."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    field_metadata: FieldMetadata = Field(alias="fieldMetadata")
    external_field_mapping_detail: ExternalFieldMapping = Field(
        alias="externalFieldMappingDetail"
    )


class EntityField(BaseModel):
    """Model representing a field within an entity."""

    model_config = ConfigDict(
        validate_by_name=True,
    )
    id: Optional[str] = Field(default=None, alias="id")
    definition: Optional[FieldMetadata] = Field(default=None, alias="definition")


class ExternalObject(BaseModel):
    """Model representing an external object."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    id: str
    external_object_name: str = Field(alias="externalObjectName")
    external_object_display_name: str = Field(alias="externalObjectDisplayName")
    primary_key: str = Field(alias="primaryKey")
    external_connection_id: str = Field(alias="externalConnectionId")
    entity_id: str = Field(alias="entityId")
    is_primary_source: bool = Field(alias="isPrimarySource")


class ExternalSourceFields(BaseModel):
    """Model representing external source fields."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    fields: List[ExternalField]
    external_object_detail: ExternalObject = Field(alias="externalObject")
    external_connection_detail: ExternalConnection = Field(alias="externalConnection")


class SourceJoinCriteria(BaseModel):
    """Model representing source join criteria."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )
    id: str
    entity_id: str = Field(alias="entityId")
    join_field_name: str = Field(alias="joinFieldName")
    join_type: str = Field(alias="joinType")
    related_source_object_id: str = Field(alias="relatedSourceObjectId")
    related_source_object_field_name: str = Field(alias="relatedSourceObjectFieldName")
    related_source_field_name: str = Field(alias="relatedSourceFieldName")


class EntityRecord(BaseModel):
    """Model representing a record within an entity."""

    model_config = {
        "validate_by_name": True,
        "validate_by_alias": True,
        "extra": "allow",
    }

    id: str = Field(alias="Id")  # Mandatory field validated by Pydantic

    @classmethod
    def from_data(
        cls, data: Dict[str, Any], model: Optional[Any] = None
    ) -> "EntityRecord":
        """Create an EntityRecord instance by validating raw data and optionally instantiating a custom model.

        :param data: Raw data dictionary for the entity.
        :param model: Optional user-defined class for validation.
        :return: EntityRecord instance
        """
        # Validate the "Id" field is mandatory and must be a string
        id_value = data.get("Id", None)
        if id_value is None or not isinstance(id_value, str):
            raise ValueError("Field 'Id' is mandatory and must be a string.")

        if model:
            # Check if the model is a plain Python class or Pydantic model
            cls._validate_against_user_model(data, model)

        return cls(**data)

    @staticmethod
    def _validate_against_user_model(
        data: Dict[str, Any], user_class: Type[Any]
    ) -> None:
        user_class_annotations = getattr(user_class, "__annotations__", None)
        if user_class_annotations is None:
            raise ValueError(
                f"User-provided class '{user_class.__name__}' is missing type annotations."
            )

        # Dynamically define a Pydantic model based on the user's class annotations
        # Fields must be valid type annotations directly
        pydantic_fields: dict[str, tuple[Any, EllipsisType | None]] = {}

        for name, annotation in user_class_annotations.items():
            is_optional = False

            origin = get_origin(annotation)
            args = get_args(annotation)

            # Handle Optional[...] or X | None
            if origin is Union and type(None) in args:
                is_optional = True

            # Check for optional fields
            if is_optional:
                pydantic_fields[name] = (annotation, None)  # Not required
            else:
                pydantic_fields[name] = (annotation, ...)

        # Dynamically create the Pydantic model class
        dynamic_model = create_model(
            f"Dynamic_{user_class.__name__}",
            **pydantic_fields,  # type: ignore[call-overload] # __base__ causes an issue. type checker cannot know that the key does not contain "__base__"
        )

        # Validate input data
        dynamic_model.model_validate(data)


class Entity(BaseModel):
    """Model representing an entity in the UiPath platform."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )

    name: str
    display_name: str = Field(alias="displayName")
    entity_type: str = Field(alias="entityType")
    description: Optional[str] = Field(default=None, alias="description")
    fields: Optional[List[FieldMetadata]] = Field(default=None, alias="fields")
    external_fields: Optional[List[ExternalSourceFields]] = Field(
        default=None, alias="externalFields"
    )
    source_join_criteria: Optional[List[SourceJoinCriteria]] = Field(
        default=None, alias="sourceJoinCriteria"
    )
    record_count: Optional[int] = Field(default=None, alias="recordCount")
    storage_size_in_mb: Optional[float] = Field(default=None, alias="storageSizeInMB")
    used_storage_size_in_mb: Optional[float] = Field(
        default=None, alias="usedStorageSizeInMB"
    )
    attachment_size_in_byte: Optional[int] = Field(
        default=None, alias="attachmentSizeInBytes"
    )
    is_rbac_enabled: bool = Field(alias="isRbacEnabled")
    id: str


class EntityRecordsBatchResponse(BaseModel):
    """Model representing a batch response of entity records."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )

    success_records: List[EntityRecord] = Field(alias="successRecords")
    failure_records: List[EntityRecord] = Field(alias="failureRecords")


Entity.model_rebuild()



"""UiPath Chat Services.

This module provides services for chat-related functionality including:
- LLM Gateway services for chat completions and embeddings
- Conversations service for autopilot conversations
"""

from ._conversations_service import ConversationsService
from ._llm_gateway_service import (
    DEFAULT_REQUESTING_FEATURE,
    DEFAULT_REQUESTING_PRODUCT,
    ChatModels,
    EmbeddingModels,
    UiPathLlmChatService,
    UiPathOpenAIService,
)
from .llm_gateway import (
    AutoToolChoice,
    ChatCompletion,
    ChatCompletionChoice,
    ChatCompletionUsage,
    ChatMessage,
    EmbeddingItem,
    EmbeddingUsage,
    RequiredToolChoice,
    SpecificToolChoice,
    TextEmbedding,
    ToolCall,
    ToolChoice,
    ToolDefinition,
    ToolFunctionDefinition,
    ToolParametersDefinition,
    ToolPropertyDefinition,
)
from .llm_throttle import get_llm_semaphore, set_llm_concurrency

__all__ = [
    # Conversations Service
    "ConversationsService",
    # LLM Gateway Services
    "ChatModels",
    "EmbeddingModels",
    "DEFAULT_REQUESTING_PRODUCT",
    "DEFAULT_REQUESTING_FEATURE",
    "UiPathLlmChatService",
    "UiPathOpenAIService",
    # LLM Throttling
    "get_llm_semaphore",
    "set_llm_concurrency",
    # LLM Gateway Models
    "ToolPropertyDefinition",
    "ToolParametersDefinition",
    "ToolFunctionDefinition",
    "ToolDefinition",
    "AutoToolChoice",
    "RequiredToolChoice",
    "SpecificToolChoice",
    "ChatMessage",
    "ChatCompletionChoice",
    "ChatCompletionUsage",
    "ChatCompletion",
    "EmbeddingItem",
    "EmbeddingUsage",
    "TextEmbedding",
    "ToolChoice",
    "ToolCall",
]



from typing import Any, Dict, List, Union

from httpx import Response
from uipath.core.tracing import traced

from ..common._base_service import BaseService
from ..common._config import UiPathApiConfig
from ..common._execution_context import UiPathExecutionContext
from ..common._folder_context import FolderContext
from ..common._models import Endpoint, RequestSpec
from .queues import (
    CommitType,
    QueueItem,
    TransactionItem,
    TransactionItemResult,
)


class QueuesService(FolderContext, BaseService):
    """Service for managing UiPath queues and queue items.

    Queues are a fundamental component of UiPath automation that enable distributed
    and scalable processing of work items.
    """

    def __init__(
        self, config: UiPathApiConfig, execution_context: UiPathExecutionContext
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="queues_list_items", run_type="uipath")
    def list_items(self) -> Response:
        """Retrieves a list of queue items from the Orchestrator.

        Returns:
            Response: HTTP response containing the list of queue items.
        """
        spec = self._list_items_spec()
        response = self.request(spec.method, url=spec.endpoint)

        return response.json()

    @traced(name="queues_list_items", run_type="uipath")
    async def list_items_async(self) -> Response:
        """Asynchronously retrieves a list of queue items from the Orchestrator.

        Returns:
            Response: HTTP response containing the list of queue items.
        """
        spec = self._list_items_spec()
        response = await self.request_async(spec.method, url=spec.endpoint)
        return response.json()

    @traced(name="queues_create_item", run_type="uipath")
    def create_item(self, item: Union[Dict[str, Any], QueueItem]) -> Response:
        """Creates a new queue item in the Orchestrator.

        Args:
            item: Queue item data, either as a dictionary or QueueItem instance.

        Returns:
            Response: HTTP response containing the created queue item details.

        Related Activity: [Add Queue Item](https://docs.uipath.com/ACTIVITIES/other/latest/workflow/add-queue-item)
        """
        spec = self._create_item_spec(item)
        response = self.request(spec.method, url=spec.endpoint, json=spec.json)
        return response.json()

    @traced(name="queues_create_item", run_type="uipath")
    async def create_item_async(
        self, item: Union[Dict[str, Any], QueueItem]
    ) -> Response:
        """Asynchronously creates a new queue item in the Orchestrator.

        Args:
            item: Queue item data, either as a dictionary or QueueItem instance.

        Returns:
            Response: HTTP response containing the created queue item details.

        Related Activity: [Add Queue Item](https://docs.uipath.com/ACTIVITIES/other/latest/workflow/add-queue-item)
        """
        spec = self._create_item_spec(item)
        response = await self.request_async(
            spec.method, url=spec.endpoint, json=spec.json
        )
        return response.json()

    @traced(name="queues_create_items", run_type="uipath")
    def create_items(
        self,
        items: List[Union[Dict[str, Any], QueueItem]],
        queue_name: str,
        commit_type: CommitType,
    ) -> Response:
        """Creates multiple queue items in bulk.

        Args:
            items: List of queue items to create, each either a dictionary or QueueItem instance.
            queue_name: Name of the target queue.
            commit_type: Type of commit operation to use for the bulk operation.

        Returns:
            Response: HTTP response containing the bulk operation result.
        """
        spec = self._create_items_spec(items, queue_name, commit_type)
        response = self.request(spec.method, url=spec.endpoint, json=spec.json)
        return response.json()

    @traced(name="queues_create_items", run_type="uipath")
    async def create_items_async(
        self,
        items: List[Union[Dict[str, Any], QueueItem]],
        queue_name: str,
        commit_type: CommitType,
    ) -> Response:
        """Asynchronously creates multiple queue items in bulk.

        Args:
            items: List of queue items to create, each either a dictionary or QueueItem instance.
            queue_name: Name of the target queue.
            commit_type: Type of commit operation to use for the bulk operation.

        Returns:
            Response: HTTP response containing the bulk operation result.
        """
        spec = self._create_items_spec(items, queue_name, commit_type)
        response = await self.request_async(
            spec.method, url=spec.endpoint, json=spec.json
        )
        return response.json()

    @traced(name="queues_create_transaction_item", run_type="uipath")
    def create_transaction_item(
        self, item: Union[Dict[str, Any], TransactionItem], no_robot: bool = False
    ) -> Response:
        """Creates a new transaction item in a queue.

        Args:
            item: Transaction item data, either as a dictionary or TransactionItem instance.
            no_robot: If True, the transaction will not be associated with a robot. Defaults to False.

        Returns:
            Response: HTTP response containing the transaction item details.
        """
        spec = self._create_transaction_item_spec(item, no_robot)
        response = self.request(spec.method, url=spec.endpoint, json=spec.json)
        return response.json()

    @traced(name="queues_create_transaction_item", run_type="uipath")
    async def create_transaction_item_async(
        self, item: Union[Dict[str, Any], TransactionItem], no_robot: bool = False
    ) -> Response:
        """Asynchronously creates a new transaction item in a queue.

        Args:
            item: Transaction item data, either as a dictionary or TransactionItem instance.
            no_robot: If True, the transaction will not be associated with a robot. Defaults to False.

        Returns:
            Response: HTTP response containing the transaction item details.
        """
        spec = self._create_transaction_item_spec(item, no_robot)
        response = await self.request_async(
            spec.method, url=spec.endpoint, json=spec.json
        )
        return response.json()

    @traced(name="queues_update_progress_of_transaction_item", run_type="uipath")
    def update_progress_of_transaction_item(
        self, transaction_key: str, progress: str
    ) -> Response:
        """Updates the progress of a transaction item.

        Args:
            transaction_key: Unique identifier of the transaction.
            progress: Progress message to set.

        Returns:
            Response: HTTP response confirming the progress update.

        Related Activity: [Set Transaction Progress](https://docs.uipath.com/activities/other/latest/workflow/set-transaction-progress)
        """
        spec = self._update_progress_of_transaction_item_spec(transaction_key, progress)
        response = self.request(spec.method, url=spec.endpoint, json=spec.json)
        return response.json()

    @traced(name="queues_update_progress_of_transaction_item", run_type="uipath")
    async def update_progress_of_transaction_item_async(
        self, transaction_key: str, progress: str
    ) -> Response:
        """Asynchronously updates the progress of a transaction item.

        Args:
            transaction_key: Unique identifier of the transaction.
            progress: Progress message to set.

        Returns:
            Response: HTTP response confirming the progress update.

        Related Activity: [Set Transaction Progress](https://docs.uipath.com/activities/other/latest/workflow/set-transaction-progress)
        """
        spec = self._update_progress_of_transaction_item_spec(transaction_key, progress)
        response = await self.request_async(
            spec.method, url=spec.endpoint, json=spec.json
        )
        return response.json()

    @traced(name="queues_complete_transaction_item", run_type="uipath")
    def complete_transaction_item(
        self, transaction_key: str, result: Union[Dict[str, Any], TransactionItemResult]
    ) -> Response:
        """Completes a transaction item with the specified result.

        Args:
            transaction_key: Unique identifier of the transaction to complete.
            result: Result data for the transaction, either as a dictionary or TransactionItemResult instance.

        Returns:
            Response: HTTP response confirming the transaction completion.

        Related Activity: [Set Transaction Status](https://docs.uipath.com/activities/other/latest/workflow/set-transaction-status)
        """
        spec = self._complete_transaction_item_spec(transaction_key, result)
        response = self.request(spec.method, url=spec.endpoint, json=spec.json)
        return response.json()

    @traced(name="queues_complete_transaction_item", run_type="uipath")
    async def complete_transaction_item_async(
        self, transaction_key: str, result: Union[Dict[str, Any], TransactionItemResult]
    ) -> Response:
        """Asynchronously completes a transaction item with the specified result.

        Args:
            transaction_key: Unique identifier of the transaction to complete.
            result: Result data for the transaction, either as a dictionary or TransactionItemResult instance.

        Returns:
            Response: HTTP response confirming the transaction completion.

        Related Activity: [Set Transaction Status](https://docs.uipath.com/activities/other/latest/workflow/set-transaction-status)
        """
        spec = self._complete_transaction_item_spec(transaction_key, result)
        response = await self.request_async(
            spec.method, url=spec.endpoint, json=spec.json
        )
        return response.json()

    @property
    def custom_headers(self) -> Dict[str, str]:
        return self.folder_headers

    def _list_items_spec(self) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint("/orchestrator_/odata/QueueItems"),
        )

    def _create_item_spec(self, item: Union[Dict[str, Any], QueueItem]) -> RequestSpec:
        if isinstance(item, dict):
            queue_item = QueueItem(**item)
        elif isinstance(item, QueueItem):
            queue_item = item

        json_payload = {
            "itemData": queue_item.model_dump(exclude_unset=True, by_alias=True)
        }

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Queues/UiPathODataSvc.AddQueueItem"
            ),
            json=json_payload,
        )

    def _create_items_spec(
        self,
        items: List[Union[Dict[str, Any], QueueItem]],
        queue_name: str,
        commit_type: CommitType,
    ) -> RequestSpec:
        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Queues/UiPathODataSvc.BulkAddQueueItems"
            ),
            json={
                "queueName": queue_name,
                "commitType": commit_type.value,
                "queueItems": [
                    item.model_dump(exclude_unset=True, by_alias=True)
                    if isinstance(item, QueueItem)
                    else QueueItem(**item).model_dump(exclude_unset=True, by_alias=True)
                    for item in items
                ],
            },
        )

    def _create_transaction_item_spec(
        self, item: Union[Dict[str, Any], TransactionItem], no_robot: bool = False
    ) -> RequestSpec:
        if isinstance(item, dict):
            transaction_item = TransactionItem(**item)
        elif isinstance(item, TransactionItem):
            transaction_item = item

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Queues/UiPathODataSvc.StartTransaction"
            ),
            json={
                "transactionData": {
                    **transaction_item.model_dump(exclude_unset=True, by_alias=True),
                    **(
                        {"RobotIdentifier": self._execution_context.robot_key}
                        if not no_robot
                        else {}
                    ),
                }
            },
        )

    def _update_progress_of_transaction_item_spec(
        self, transaction_key: str, progress: str
    ) -> RequestSpec:
        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                f"/orchestrator_/odata/QueueItems({transaction_key})/UiPathODataSvc.SetTransactionProgress"
            ),
            json={"progress": progress},
        )

    def _complete_transaction_item_spec(
        self, transaction_key: str, result: Union[Dict[str, Any], TransactionItemResult]
    ) -> RequestSpec:
        if isinstance(result, dict):
            transaction_result = TransactionItemResult(**result)
        elif isinstance(result, TransactionItemResult):
            transaction_result = result

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                f"/orchestrator_/odata/Queues({transaction_key})/UiPathODataSvc.SetTransactionResult"
            ),
            json={
                "transactionResult": transaction_result.model_dump(
                    exclude_unset=True, by_alias=True
                )
            },
        )



"""Module defining the ActionSchema model for UiPath platform actions."""

from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field


class FieldDetails(BaseModel):
    """Model representing details of a field in an action schema."""

    name: str
    key: str


class TaskSchema(BaseModel):
    """Model representing the schema of an action in the UiPath platform."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )

    key: str
    in_outs: Optional[List[FieldDetails]] = Field(default=None, alias="inOuts")
    inputs: Optional[List[FieldDetails]] = None
    outputs: Optional[List[FieldDetails]] = None
    outcomes: Optional[List[FieldDetails]] = None



"""Models for UiPath Portal service."""

from pydantic import BaseModel


class TenantInfo(BaseModel):
    """Model representing a tenant."""

    name: str
    id: str


class OrganizationInfo(BaseModel):
    """Model representing an organization."""

    id: str
    name: str


class TenantsAndOrganizationInfoResponse(BaseModel):
    """Model representing the tenants and organization info response."""

    tenants: list[TenantInfo]
    organization: OrganizationInfo



"""UiPath Entities Models.

This module contains models related to UiPath Entities service.
"""

from ._entities_service import EntitiesService
from .entities import (
    Entity,
    EntityField,
    EntityFieldMetadata,
    EntityRecord,
    EntityRecordsBatchResponse,
    ExternalField,
    ExternalObject,
    ExternalSourceFields,
    FieldDataType,
    FieldMetadata,
    ReferenceType,
    SourceJoinCriteria,
)

__all__ = [
    "EntitiesService",
    "Entity",
    "EntityField",
    "EntityRecord",
    "EntityFieldMetadata",
    "FieldDataType",
    "FieldMetadata",
    "EntityRecordsBatchResponse",
    "ExternalField",
    "ExternalObject",
    "ExternalSourceFields",
    "ReferenceType",
    "SourceJoinCriteria",
]



import json

import httpx

from uipath.platform.errors import EnrichedException

# URL patterns that route to specific extractors
_ORCHESTRATOR_URL = "https://cloud.uipath.com/org/tenant/orchestrator_/api/v1"
_LLM_URL = "https://cloud.uipath.com/org/tenant/orchestrator_/llm/chat"
_AGENTHUB_URL = "https://cloud.uipath.com/org/tenant/agenthub_/api/v1"
_APPS_URL = "https://cloud.uipath.com/org/tenant/apps_/api/v1"
_ELEMENTS_URL = "https://cloud.uipath.com/org/tenant/elements_/api/v1"
_LLMOPS_URL = "https://cloud.uipath.com/org/tenant/llmopstenant_/api/v1"
_GENERIC_URL = "https://cloud.uipath.com/api/test"


def _make_error(
    status_code: int = 500,
    body: str | bytes = b"",
    content_type: str = "application/json",
    url: str = _GENERIC_URL,
    method: str = "POST",
) -> httpx.HTTPStatusError:
    if isinstance(body, str):
        body = body.encode()
    return httpx.HTTPStatusError(
        message=f"Server error {status_code}",
        request=httpx.Request(method, url),
        response=httpx.Response(
            status_code,
            content=body,
            headers={"content-type": content_type},
        ),
    )


class TestBasicProperties:
    def test_status_code(self):
        exc = EnrichedException(_make_error(404))
        assert exc.status_code == 404

    def test_url(self):
        exc = EnrichedException(
            _make_error(url="https://cloud.uipath.com/org/tenant/api/v1")
        )
        assert exc.url == "https://cloud.uipath.com/org/tenant/api/v1"

    def test_http_method(self):
        exc = EnrichedException(_make_error(method="DELETE"))
        assert exc.http_method == "DELETE"

    def test_response_content_short(self):
        exc = EnrichedException(_make_error(body=b'{"message":"oops"}'))
        assert exc.response_content == '{"message":"oops"}'

    def test_response_content_truncated(self):
        long_body = "x" * 300
        exc = EnrichedException(_make_error(body=long_body.encode()))
        assert exc.response_content.endswith("... (truncated)")
        assert len(exc.response_content) < 300

    def test_str_contains_status_code(self):
        exc = EnrichedException(_make_error(404))
        assert "404" in str(exc)

    def test_status_code_is_int(self):
        exc = EnrichedException(_make_error(502))
        assert isinstance(exc.status_code, int)


class TestErrorInfo:
    def test_none_for_empty_body(self):
        exc = EnrichedException(_make_error(body=b""))
        assert exc.error_info is None

    def test_none_for_non_json(self):
        exc = EnrichedException(_make_error(body=b"not json"))
        assert exc.error_info is None

    def test_cached(self):
        body = json.dumps({"message": "test"})
        exc = EnrichedException(_make_error(body=body))
        first = exc.error_info
        second = exc.error_info
        assert first is second


class TestGenericExtraction:
    """Tests for the generic fallback extractor (no service prefix in URL)."""

    def test_message_field(self):
        body = json.dumps({"message": "some error"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.message == "some error"

    def test_error_as_string(self):
        body = json.dumps({"error": "something went wrong"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.message == "something went wrong"

    def test_nested_error_message(self):
        body = json.dumps({"error": {"message": "nested error"}})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.message == "nested error"

    def test_error_code_fields(self):
        body = json.dumps({"errorCode": "ERR_001"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.error_code == "ERR_001"

    def test_code_field(self):
        body = json.dumps({"code": "ERR_002"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.error_code == "ERR_002"

    def test_numeric_code_converted_to_str(self):
        body = json.dumps({"errorCode": 1002})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.error_code == "1002"

    def test_trace_id(self):
        body = json.dumps({"traceId": "abc-123"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.trace_id == "abc-123"

    def test_request_id(self):
        body = json.dumps({"requestId": "req-456"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.trace_id == "req-456"

    def test_no_known_fields_returns_none_message(self):
        body = json.dumps({"unknown_field": "value"})
        exc = EnrichedException(_make_error(body=body))
        assert exc.error_info is not None
        assert exc.error_info.message is None


class TestRfc7807Extraction:
    """Tests for RFC 7807 content-type based extraction."""

    def test_problem_json_content_type(self):
        body = json.dumps(
            {"detail": "Forbidden", "title": "Access Denied", "type": "AUTH_ERR"}
        )
        exc = EnrichedException(
            _make_error(403, body=body, content_type="application/problem+json")
        )
        assert exc.error_info is not None
        assert exc.error_info.message == "Forbidden"
        assert exc.error_info.error_code == "AUTH_ERR"

    def test_structural_rfc7807_detection(self):
        body = json.dumps(
            {"detail": "Not found", "title": "Error", "status": 404, "type": "E1"}
        )
        exc = EnrichedException(_make_error(404, body=body))
        assert exc.error_info is not None
        assert exc.error_info.message == "Not found"
        assert exc.error_info.error_code == "E1"


class TestOrchestratorExtraction:
    """Orchestrator uses generic extractor (message, errorCode, traceId)."""

    def test_process_not_found(self):
        body = json.dumps(
            {
                "message": "Process not found",
                "errorCode": "1002",
                "traceId": "abc-123",
            }
        )
        exc = EnrichedException(_make_error(404, body=body, url=_ORCHESTRATOR_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Process not found"
        assert exc.error_info.error_code == "1002"
        assert exc.error_info.trace_id == "abc-123"


class TestLlmGatewayExtraction:
    """LLM Gateway has nested error objects and vendor passthrough."""

    def test_nested_vendor_error(self):
        body = json.dumps(
            {
                "error": {
                    "code": "content_filter",
                    "message": "Content was filtered by safety policy",
                }
            }
        )
        exc = EnrichedException(_make_error(400, body=body, url=_LLM_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Content was filtered by safety policy"
        assert exc.error_info.error_code == "content_filter"

    def test_top_level_with_nested(self):
        body = json.dumps(
            {
                "message": "Model unavailable",
                "errorCode": "LLM_ERR",
                "error": {"message": "rate limit", "code": "429"},
            }
        )
        exc = EnrichedException(_make_error(429, body=body, url=_LLM_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Model unavailable"
        assert exc.error_info.error_code == "LLM_ERR"


class TestAgentHubExtraction:
    """AgentHub has nested problem objects."""

    def test_nested_problem(self):
        body = json.dumps({"problem": {"detail": "Agent not found", "type": "AH_404"}})
        exc = EnrichedException(_make_error(404, body=body, url=_AGENTHUB_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Agent not found"
        assert exc.error_info.error_code == "AH_404"


class TestAppsExtraction:
    """Apps prioritizes errorMessageV2 over message."""

    def test_error_message_v2_priority(self):
        body = json.dumps(
            {"errorMessageV2": "Detailed error", "message": "Short error"}
        )
        exc = EnrichedException(_make_error(400, body=body, url=_APPS_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Detailed error"


class TestElementsExtraction:
    """Elements has vendor passthrough fields (providerMessage, providerErrorCode)."""

    def test_provider_fields(self):
        body = json.dumps(
            {
                "providerMessage": "Rate limit exceeded",
                "providerErrorCode": "429",
                "requestId": "req-123",
            }
        )
        exc = EnrichedException(_make_error(429, body=body, url=_ELEMENTS_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Rate limit exceeded"
        assert exc.error_info.error_code == "429"
        assert exc.error_info.trace_id == "req-123"


class TestLlmOpsExtraction:
    """LLMOps uses errorMessage, code, requestId."""

    def test_llmops_fields(self):
        body = json.dumps(
            {
                "errorMessage": "Quota exceeded",
                "code": "QUOTA_001",
                "requestId": "req-789",
            }
        )
        exc = EnrichedException(_make_error(429, body=body, url=_LLMOPS_URL))
        assert exc.error_info is not None
        assert exc.error_info.message == "Quota exceeded"
        assert exc.error_info.error_code == "QUOTA_001"
        assert exc.error_info.trace_id == "req-789"


class TestRouterPriority:
    """RFC 7807 content-type should override service-specific extractors."""

    def test_rfc7807_content_type_beats_service_extractor(self):
        body = json.dumps(
            {"detail": "RFC detail", "type": "RFC_ERR", "errorMessageV2": "Apps msg"}
        )
        exc = EnrichedException(
            _make_error(
                400, body=body, url=_APPS_URL, content_type="application/problem+json"
            )
        )
        assert exc.error_info is not None
        assert exc.error_info.message == "RFC detail"
        assert exc.error_info.error_code == "RFC_ERR"



"""Elements service error payload extractor."""

from typing import Any

from .._enriched_exception import ExtractedErrorInfo
from ._helpers import get_str_field, get_typed_field


def extract_elements(body: dict[str, Any]) -> ExtractedErrorInfo:
    message = get_typed_field(body, str, "providerMessage") or get_typed_field(
        body, str, "message"
    )
    error_code = get_str_field(body, "providerErrorCode", "status")
    trace_id = get_typed_field(body, str, "requestId")

    return ExtractedErrorInfo(
        message=message,
        error_code=error_code,
        trace_id=trace_id,
    )



"""Utility for retrieving the Orchestrator server version."""

import httpx

from ..common._http_config import get_httpx_client_kwargs


def get_server_version(domain: str) -> str | None:
    """Get the Orchestrator server version.

    Args:
        domain: The base URL of the UiPath platform (e.g., "https://cloud.uipath.com").

    Returns:
        The server version string, or None if the request fails for any reason.
    """
    url = f"{domain}/orchestrator_/api/status/version"

    try:
        client_kwargs = get_httpx_client_kwargs()
        client_kwargs["timeout"] = 5.0
        with httpx.Client(**client_kwargs) as client:
            response = client.get(url)
            response.raise_for_status()
            data = response.json()
            return data.get("version")
    except Exception:
        return None


async def get_server_version_async(domain: str) -> str | None:
    """Get the Orchestrator server version.

    Args:
        domain: The base URL of the UiPath platform (e.g., "https://cloud.uipath.com").

    Returns:
        The server version string, or None if the request fails for any reason.
    """
    url = f"{domain}/orchestrator_/api/status/version"

    try:
        client_kwargs = get_httpx_client_kwargs()
        client_kwargs["timeout"] = 5.0
        async with httpx.AsyncClient(**client_kwargs) as client:
            response = await client.get(url)
            response.raise_for_status()
            data = response.json()
            return data.get("version")
    except Exception:
        return None



"""Identity service for UiPath authentication token operations."""

from typing import Optional

import httpx

from ..common._http_config import get_httpx_client_kwargs
from ..common.auth import TokenData


class IdentityService:
    """Service for interacting with the UiPath Identity server."""

    def __init__(self, domain: str):
        """Initialize the IdentityService.

        Args:
            domain: The base URL of the UiPath identity server (e.g., "https://cloud.uipath.com").
        """
        self._domain = domain

    def refresh_access_token(
        self,
        refresh_token: str,
        client_id: str,
    ) -> TokenData:
        """Refresh an access token using a refresh token.

        Args:
            refresh_token: The refresh token to exchange for a new access token.
            client_id: The client ID of the application.

        Returns:
            TokenData containing the new access token and related information.

        Raises:
            httpx.HTTPStatusError: If the server returns a non-2xx response.
            httpx.ConnectError: If there is a network connectivity issue.
        """
        url = f"{self._domain}/identity_/connect/token"
        data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": client_id,
        }

        with httpx.Client(**get_httpx_client_kwargs()) as client:
            response = client.post(url, data=data)
            response.raise_for_status()
            return TokenData.model_validate(response.json())

    async def refresh_access_token_async(
        self,
        refresh_token: str,
        client_id: str,
    ) -> TokenData:
        """Refresh an access token using a refresh token.

        Args:
            refresh_token: The refresh token to exchange for a new access token.
            client_id: The client ID of the application.

        Returns:
            TokenData containing the new access token and related information.

        Raises:
            httpx.HTTPStatusError: If the server returns a non-2xx response.
            httpx.ConnectError: If there is a network connectivity issue.
        """
        url = f"{self._domain}/identity_/connect/token"
        data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": client_id,
        }

        async with httpx.AsyncClient(**get_httpx_client_kwargs()) as client:
            response = await client.post(url, data=data)
            response.raise_for_status()
            return TokenData.model_validate(response.json())

    def get_client_credentials_token(
        self,
        client_id: str,
        client_secret: str,
        scope: Optional[str] = "OR.Execution",
    ) -> TokenData:
        """Obtain an access token using client credentials grant.

        Args:
            client_id: The client ID of the application.
            client_secret: The client secret of the application.
            scope: The requested OAuth scopes (optional, default: "OR.Execution").

        Returns:
            TokenData containing the access token and related information.

        Raises:
            httpx.HTTPStatusError: If the server returns a non-2xx response.
            httpx.ConnectError: If there is a network connectivity issue.
        """
        scope = scope or "OR.Execution"
        url = f"{self._domain}/identity_/connect/token"
        data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": scope,
        }

        with httpx.Client(**get_httpx_client_kwargs()) as client:
            response = client.post(url, data=data)
            response.raise_for_status()
            return TokenData.model_validate(response.json())

    async def get_client_credentials_token_async(
        self,
        client_id: str,
        client_secret: str,
        scope: Optional[str] = "OR.Execution",
    ) -> TokenData:
        """Obtain an access token using client credentials grant.

        Args:
            client_id: The client ID of the application.
            client_secret: The client secret of the application.
            scope: The requested OAuth scopes (optional, default: "OR.Execution").

        Returns:
            TokenData containing the access token and related information.

        Raises:
            httpx.HTTPStatusError: If the server returns a non-2xx response.
            httpx.ConnectError: If there is a network connectivity issue.
        """
        scope = scope or "OR.Execution"
        url = f"{self._domain}/identity_/connect/token"
        data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": scope,
        }

        async with httpx.AsyncClient(**get_httpx_client_kwargs()) as client:
            response = await client.post(url, data=data)
            response.raise_for_status()
            return TokenData.model_validate(response.json())



"""AgentHub / AgentsRuntime error payload extractor."""

from typing import Any

from .._enriched_exception import ExtractedErrorInfo
from ._helpers import get_typed_field


def extract_agenthub(body: dict[str, Any]) -> ExtractedErrorInfo:
    message = get_typed_field(body, str, "message", "title")
    error_code = get_typed_field(body, str, "errorCode", "code")
    trace_id = get_typed_field(body, str, "traceId")

    problem = get_typed_field(body, dict, "problem")
    if problem is not None:
        message = message or get_typed_field(problem, str, "detail", "title")
        error_code = error_code or get_typed_field(problem, str, "errorCode", "type")
        trace_id = trace_id or get_typed_field(problem, str, "traceId")

    return ExtractedErrorInfo(
        message=message,
        error_code=error_code,
        trace_id=trace_id,
    )



"""UiPath Portal Service."""

from ._portal_service import PortalService
from .portal import OrganizationInfo, TenantInfo, TenantsAndOrganizationInfoResponse

__all__ = [
    "PortalService",
    "TenantInfo",
    "OrganizationInfo",
    "TenantsAndOrganizationInfoResponse",
]



"""UiPath resume trigger enums."""

from enum import Enum
from typing import Any

from pydantic import BaseModel, Field


class PropertyName(str, Enum):
    """UiPath trigger property names."""

    INTERNAL = "__internal"


class TriggerMarker(str, Enum):
    """UiPath trigger markers.

    These markers are used as properties of resume triggers objects for special handling at runtime.
    """

    NO_CONTENT = "NO_CONTENT"


def is_no_content_marker(value: Any) -> bool:
    """Check if a value is a NO_CONTENT trigger marker (dict or string form)."""
    if isinstance(value, dict):
        return value.get(PropertyName.INTERNAL.value) == TriggerMarker.NO_CONTENT.value
    if isinstance(value, str):
        return (
            PropertyName.INTERNAL.value in value
            and TriggerMarker.NO_CONTENT.value in value
        )
    return False


class ExternalTriggerType(str, Enum):
    """External trigger types."""

    DEEP_RAG = "deepRag"
    BATCH_RAG = "batchRag"
    IXP_EXTRACTION = "ixpExtraction"
    INDEX_INGESTION = "indexIngestion"
    IXP_VS_ESCALATION = "IxpVsEscalation"


class ExternalTrigger(BaseModel):
    """Model representing an external trigger entity."""

    type: ExternalTriggerType
    external_id: str = Field(alias="externalId")

    model_config = {
        "validate_by_name": True,
        "validate_by_alias": True,
    }



class OperationFailedException(Exception):
    """Raised when attempting to get results from a failed operation.

    This exception is raised when attempting to retrieve results from operation
    that failed to complete successfully.
    """

    def __init__(
        self,
        operation_id: str,
        status: str,
        error: str,
        operation_name: str = "Operation",
    ):
        self.operation_id = operation_id
        self.status = status
        self.error = error
        self.message = f"{operation_name} '{operation_id}' failed with status: {status} error: {error}"
        super().__init__(self.message)



"""Implementation of UiPath resume trigger protocols."""

import json
import os
import uuid
from typing import Any

from uipath.core.errors import (
    ErrorCategory,
    UiPathFaultedTriggerError,
    UiPathPendingTriggerError,
)
from uipath.core.serialization import serialize_object
from uipath.core.triggers import (
    UiPathApiTrigger,
    UiPathResumeTrigger,
    UiPathResumeTriggerName,
    UiPathResumeTriggerType,
)

from uipath.platform import UiPath
from uipath.platform.action_center import Task
from uipath.platform.action_center.tasks import TaskStatus
from uipath.platform.common._config import UiPathConfig
from uipath.platform.common.interrupt_models import (
    CreateBatchTransform,
    CreateDeepRag,
    CreateDeepRagRaw,
    CreateEphemeralIndex,
    CreateEphemeralIndexRaw,
    CreateEscalation,
    CreateTask,
    DocumentExtraction,
    DocumentExtractionValidation,
    InvokeProcess,
    InvokeProcessRaw,
    InvokeSystemAgent,
    WaitBatchTransform,
    WaitDeepRag,
    WaitDeepRagRaw,
    WaitDocumentExtraction,
    WaitDocumentExtractionValidation,
    WaitEphemeralIndex,
    WaitEphemeralIndexRaw,
    WaitEscalation,
    WaitJob,
    WaitJobRaw,
    WaitSystemAgent,
    WaitTask,
)
from uipath.platform.context_grounding import DeepRagStatus, IndexStatus
from uipath.platform.context_grounding.context_grounding_index import (
    ContextGroundingIndex,
)
from uipath.platform.errors import (
    BatchTransformNotCompleteException,
    OperationNotCompleteException,
)
from uipath.platform.orchestrator.job import JobState
from uipath.platform.resume_triggers._enums import (
    ExternalTrigger,
    ExternalTriggerType,
    PropertyName,
    TriggerMarker,
)


def _try_convert_to_json_format(value: str | None) -> Any:
    """Attempts to parse a string as JSON and returns the parsed object or original string.

    Args:
        value: The string value to attempt JSON parsing on.

    Returns:
        The parsed JSON object if successful, otherwise the original string value.
    """
    try:
        if not value:
            return None
        return json.loads(value)
    except json.decoder.JSONDecodeError:
        return value


class UiPathResumeTriggerReader:
    """Handles reading and retrieving Human-In-The-Loop (HITL) data from UiPath services.

    Implements UiPathResumeTriggerReaderProtocol.
    """

    def _extract_field(self, field_name: str, payload: Any) -> str | None:
        """Extracts a field from the payload and returns it if it exists."""
        if not payload:
            return payload

        if isinstance(payload, dict):
            return payload.get(field_name)

        # 2.3.0 remove
        try:
            payload_dict = json.loads(payload)
            return payload_dict.get(field_name)
        except json.decoder.JSONDecodeError:
            return None

    async def read_trigger(self, trigger: UiPathResumeTrigger) -> Any | None:
        """Read a resume trigger and convert it to runtime-compatible input.

        This method retrieves data from UiPath services (Actions, Jobs, API)
        based on the trigger type and returns it in a format that the
        runtime can use to resume execution.

        Args:
            trigger: The resume trigger to read

        Returns:
            The data retrieved from UiPath services, ready to be used
            as resume input. Format depends on trigger type:
            - TASK: Task data (possibly with escalation processing)
            - JOB: Job output data
            - API: API payload
            Returns None if no data is available.

        Raises:
            UiPathRuntimeError: If reading fails, job failed, API connection failed,
                trigger type is unknown, or HITL feedback retrieval failed.
        """
        uipath = UiPath()

        match trigger.trigger_type:
            case UiPathResumeTriggerType.TASK:
                if trigger.item_key:
                    task: Task = await uipath.tasks.retrieve_async(
                        trigger.item_key,
                        app_folder_key=trigger.folder_key,
                        app_folder_path=trigger.folder_path,
                        app_name=self._extract_field("app_name", trigger.payload),
                    )
                    pending_status = TaskStatus.PENDING.value
                    unassigned_status = TaskStatus.UNASSIGNED.value

                    if (
                        task.status in (pending_status, unassigned_status)
                        and not task.is_deleted
                    ):
                        # 2.3.0 remove (task.status will already be the enum)
                        current_status = (
                            TaskStatus(task.status).name
                            if isinstance(task.status, int)
                            else "Unknown"
                        )
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"Task is not completed yet. Current status: {current_status}",
                        )

                    if trigger.trigger_name == UiPathResumeTriggerName.ESCALATION:
                        return task

                    # if task is deleted, raise error only for Task models
                    # for escalations we return the task object to allow graceful handling down the stream
                    if task.is_deleted:
                        raise UiPathFaultedTriggerError(
                            ErrorCategory.USER,
                            f"The {task.title} task was deleted",
                        )

                    trigger_response = task.data
                    if not bool(trigger_response):
                        # 2.3.0 change to task.status.name
                        assert isinstance(task.status, int)
                        trigger_response = {
                            "status": TaskStatus(task.status).name.lower(),
                            PropertyName.INTERNAL.value: TriggerMarker.NO_CONTENT.value,
                        }

                    return trigger_response

            case UiPathResumeTriggerType.JOB:
                if trigger.item_key:
                    job = await uipath.jobs.retrieve_async(
                        trigger.item_key,
                        folder_key=trigger.folder_key,
                        folder_path=trigger.folder_path,
                        process_name=self._extract_field("name", trigger.payload),
                    )

                    job_state = (job.state or "").lower()
                    successful_state = JobState.SUCCESSFUL.value
                    faulted_state = JobState.FAULTED.value
                    running_state = JobState.RUNNING.value
                    pending_state = JobState.PENDING.value
                    resumed_state = JobState.RESUMED.value
                    suspended_state = JobState.SUSPENDED.value

                    if job_state in (
                        pending_state,
                        running_state,
                        suspended_state,
                        resumed_state,
                    ):
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"Job is not finished yet. Current state: {job_state}",
                        )

                    if trigger.trigger_name == UiPathResumeTriggerName.JOB_RAW:
                        return job

                    if job_state != successful_state:
                        job_error = (
                            _try_convert_to_json_format(str(job.job_error or job.info))
                            or "Job error unavailable."
                            if job_state == faulted_state
                            else f"Job {job.key} is {job_state}."
                        )
                        raise UiPathFaultedTriggerError(
                            ErrorCategory.USER,
                            f"Process did not finish successfully. Error: {job_error}",
                        )

                    output_data = await uipath.jobs.extract_output_async(job)
                    trigger_response = _try_convert_to_json_format(output_data)

                    # if response is an empty dictionary, use job state as placeholder value
                    if isinstance(trigger_response, dict) and not bool(
                        trigger_response
                    ):
                        # 2.3.0 change to job_state.value
                        trigger_response = {
                            "state": job_state,
                            PropertyName.INTERNAL.value: TriggerMarker.NO_CONTENT.value,
                        }

                    return trigger_response
            case UiPathResumeTriggerType.DEEP_RAG:
                if trigger.item_key:
                    deep_rag = await uipath.context_grounding.retrieve_deep_rag_async(
                        trigger.item_key,
                        index_name=self._extract_field("index_name", trigger.payload),
                    )
                    deep_rag_status = deep_rag.last_deep_rag_status

                    if deep_rag_status in (
                        DeepRagStatus.QUEUED,
                        DeepRagStatus.IN_PROGRESS,
                    ):
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"DeepRag is not finished yet. Current status: {deep_rag_status}",
                        )

                    if trigger.trigger_name == UiPathResumeTriggerName.DEEP_RAG_RAW:
                        return deep_rag

                    if deep_rag_status != DeepRagStatus.SUCCESSFUL:
                        raise UiPathFaultedTriggerError(
                            ErrorCategory.USER,
                            f"DeepRag '{deep_rag.name}' did not finish successfully.",
                        )

                    trigger_response = deep_rag.content

                    # if response is an empty dictionary, use Deep Rag state as placeholder value
                    if not trigger_response:
                        trigger_response = {
                            "status": deep_rag_status,
                            PropertyName.INTERNAL.value: TriggerMarker.NO_CONTENT.value,
                        }
                    else:
                        trigger_response = trigger_response.model_dump()
                        trigger_response["deepRagId"] = trigger.item_key

                    return trigger_response

            case UiPathResumeTriggerType.INDEX_INGESTION:
                if trigger.item_key:
                    index = await uipath.context_grounding.retrieve_by_id_async(
                        trigger.item_key
                    )

                    ephemeral_index = ContextGroundingIndex(**index)

                    ephemeral_index_status = ephemeral_index.last_ingestion_status

                    if ephemeral_index_status in (
                        IndexStatus.QUEUED,
                        IndexStatus.IN_PROGRESS,
                    ):
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"Index ingestion is not finished yet. Current status: {ephemeral_index_status}",
                        )

                    if (
                        trigger.trigger_name
                        == UiPathResumeTriggerName.INDEX_INGESTION_RAW
                    ):
                        return ephemeral_index

                    if ephemeral_index_status != IndexStatus.SUCCESSFUL:
                        raise UiPathFaultedTriggerError(
                            ErrorCategory.USER,
                            f"Index ingestion '{ephemeral_index.name}' did not finish successfully.",
                        )

                    trigger_response = ephemeral_index.model_dump()

                    return trigger_response

            case UiPathResumeTriggerType.BATCH_RAG:
                if trigger.item_key:
                    destination_path = self._extract_field(
                        "destination_path", trigger.payload
                    )
                    assert destination_path is not None
                    try:
                        await uipath.context_grounding.download_batch_transform_result_async(
                            trigger.item_key,
                            destination_path,
                            validate_status=True,
                            index_name=self._extract_field(
                                "index_name", trigger.payload
                            ),
                        )
                    except BatchTransformNotCompleteException as e:
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"{e.message}",
                        ) from e

                    return f"Batch transform completed. Modified file available at {os.path.abspath(destination_path)}"

            case UiPathResumeTriggerType.IXP_EXTRACTION:
                if trigger.item_key:
                    project_id = self._extract_field("project_id", trigger.payload)
                    tag = self._extract_field("tag", trigger.payload)

                    assert project_id is not None
                    assert tag is not None

                    try:
                        extraction_response = (
                            await uipath.documents.retrieve_ixp_extraction_result_async(
                                project_id, tag, trigger.item_key
                            )
                        )
                    except OperationNotCompleteException as e:
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"{e.message}",
                        ) from e

                    return extraction_response.model_dump()

            case UiPathResumeTriggerType.IXP_VS_ESCALATION:
                if trigger.item_key:
                    project_id = self._extract_field("project_id", trigger.payload)
                    tag = self._extract_field("tag", trigger.payload)

                    assert project_id is not None
                    assert tag is not None
                    try:
                        escalation_response = await uipath.documents.retrieve_ixp_extraction_validation_result_async(
                            project_id, tag, trigger.item_key
                        )
                    except OperationNotCompleteException as e:
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"{e.message}",
                        ) from e

                    pending_status_name = TaskStatus.PENDING.name.lower()
                    unassigned_status_name = TaskStatus.UNASSIGNED.name.lower()

                    current_status = escalation_response.action_data["status"].lower()
                    if current_status in (pending_status_name, unassigned_status_name):
                        raise UiPathPendingTriggerError(
                            ErrorCategory.SYSTEM,
                            f"Document extraction escalation task is not completed yet. Current status: {current_status}",
                        )

                    return escalation_response.model_dump()

            case UiPathResumeTriggerType.API:
                if trigger.api_resume and trigger.api_resume.inbox_id:
                    try:
                        return await uipath.jobs.retrieve_api_payload_async(
                            trigger.api_resume.inbox_id
                        )
                    except Exception as e:
                        raise UiPathFaultedTriggerError(
                            ErrorCategory.SYSTEM,
                            f"Failed to get trigger payload"
                            f"Error fetching API trigger payload for inbox {trigger.api_resume.inbox_id}: {str(e)}",
                        ) from e

            case _:
                raise UiPathFaultedTriggerError(
                    ErrorCategory.SYSTEM,
                    f"Unexpected trigger type received"
                    f"Trigger type :{type(trigger.trigger_type)} is invalid",
                )

        raise UiPathFaultedTriggerError(
            ErrorCategory.SYSTEM, "Failed to receive payload from HITL action"
        )


class UiPathResumeTriggerCreator:
    """Creates resume triggers from suspend values.

    Implements UiPathResumeTriggerCreatorProtocol.
    """

    async def create_trigger(self, suspend_value: Any) -> UiPathResumeTrigger:
        """Create a resume trigger from a suspend value.

        This method processes the input value and creates an appropriate resume trigger
        for HITL scenarios. It handles different input types:
        - Tasks: Creates or references UiPath tasks with folder information
        - Jobs: Invokes processes or references existing jobs with folder information
        - API: Creates API triggers with generated inbox IDs

        Args:
            suspend_value: The value that caused the suspension.
                Can be UiPath models (CreateTask, InvokeProcess, etc.),
                strings, or any other value that needs HITL processing.

        Returns:
            UiPathResumeTrigger ready to be persisted

        Raises:
            UiPathRuntimeError: If action/job creation fails, escalation fails, or an
                unknown model type is encountered.
            Exception: If any underlying UiPath service calls fail.
        """
        try:
            trigger_type = self._determine_trigger_type(suspend_value)
            trigger_name = self._determine_trigger_name(suspend_value)

            resume_trigger = UiPathResumeTrigger(
                trigger_type=trigger_type,
                trigger_name=trigger_name,
                payload=serialize_object(suspend_value),
            )

            match trigger_type:
                case UiPathResumeTriggerType.TASK:
                    await self._handle_task_trigger(suspend_value, resume_trigger)

                case UiPathResumeTriggerType.JOB:
                    await self._handle_job_trigger(suspend_value, resume_trigger)

                case UiPathResumeTriggerType.API:
                    self._handle_api_trigger(suspend_value, resume_trigger)

                case UiPathResumeTriggerType.DEEP_RAG:
                    await self._handle_deep_rag_job_trigger(
                        suspend_value, resume_trigger
                    )
                case UiPathResumeTriggerType.INDEX_INGESTION:
                    await self._handle_ephemeral_index_job_trigger(
                        suspend_value, resume_trigger
                    )
                case UiPathResumeTriggerType.BATCH_RAG:
                    await self._handle_batch_rag_job_trigger(
                        suspend_value, resume_trigger
                    )
                case UiPathResumeTriggerType.IXP_EXTRACTION:
                    await self._handle_ixp_extraction_trigger(
                        suspend_value, resume_trigger
                    )
                case UiPathResumeTriggerType.IXP_VS_ESCALATION:
                    await self._handle_ixp_vs_escalation_trigger(
                        suspend_value, resume_trigger
                    )
                case _:
                    raise UiPathFaultedTriggerError(
                        ErrorCategory.SYSTEM,
                        f"Unexpected model received"
                        f"{type(suspend_value)} is not a valid Human-In-The-Loop model",
                    )
        except Exception as e:
            raise UiPathFaultedTriggerError(
                ErrorCategory.SYSTEM,
                "Failed to create HITL action",
                f"{str(e)}",
            ) from e
        return resume_trigger

    async def _create_external_trigger(self, external_trigger: ExternalTrigger):
        """Creates an external trigger in orchestrator."""
        # only create external trigger entities for non-debug runs
        if not UiPathConfig.job_key:
            return

        uipath = UiPath()
        await uipath.api_client.request_async(
            method="POST",
            url="orchestrator_/api/JobTriggers/SaveExternalTrigger",
            json=external_trigger.model_dump(by_alias=True),
        )

    def _determine_trigger_type(self, value: Any) -> UiPathResumeTriggerType:
        """Determines the resume trigger type based on the input value.

        Args:
            value: The suspend value to analyze

        Returns:
            The appropriate UiPathResumeTriggerType based on the input value type.
        """
        if isinstance(value, (CreateTask, WaitTask, CreateEscalation, WaitEscalation)):
            return UiPathResumeTriggerType.TASK
        if isinstance(
            value,
            (
                InvokeProcess,
                InvokeProcessRaw,
                WaitJob,
                WaitJobRaw,
                InvokeSystemAgent,
                WaitSystemAgent,
            ),
        ):
            return UiPathResumeTriggerType.JOB
        if isinstance(
            value, (CreateDeepRag, CreateDeepRagRaw, WaitDeepRag, WaitDeepRagRaw)
        ):
            return UiPathResumeTriggerType.DEEP_RAG
        if isinstance(value, (CreateEphemeralIndex, WaitEphemeralIndex)):
            return UiPathResumeTriggerType.INDEX_INGESTION
        if isinstance(value, (CreateBatchTransform, WaitBatchTransform)):
            return UiPathResumeTriggerType.BATCH_RAG
        if isinstance(value, (DocumentExtraction, WaitDocumentExtraction)):
            return UiPathResumeTriggerType.IXP_EXTRACTION
        if isinstance(
            value, (DocumentExtractionValidation, WaitDocumentExtractionValidation)
        ):
            return UiPathResumeTriggerType.IXP_VS_ESCALATION
        # default to API trigger
        return UiPathResumeTriggerType.API

    def _determine_trigger_name(self, value: Any) -> UiPathResumeTriggerName:
        """Determines the resume trigger name based on the input value.

        Args:
            value: The suspend value to analyze

        Returns:
            The appropriate UiPathResumeTriggerName based on the input value type.
        """
        if isinstance(value, (CreateEscalation, WaitEscalation)):
            return UiPathResumeTriggerName.ESCALATION
        if isinstance(value, (CreateTask, WaitTask)):
            return UiPathResumeTriggerName.TASK
        if isinstance(value, (InvokeProcessRaw, WaitJobRaw)):
            return UiPathResumeTriggerName.JOB_RAW
        if isinstance(
            value, (InvokeProcess, WaitJob, InvokeSystemAgent, WaitSystemAgent)
        ):
            return UiPathResumeTriggerName.JOB
        if isinstance(value, (CreateDeepRagRaw, WaitDeepRagRaw)):
            return UiPathResumeTriggerName.DEEP_RAG_RAW
        if isinstance(value, (CreateDeepRag, WaitDeepRag)):
            return UiPathResumeTriggerName.DEEP_RAG
        if isinstance(value, (CreateEphemeralIndexRaw, WaitEphemeralIndexRaw)):
            return UiPathResumeTriggerName.INDEX_INGESTION_RAW
        if isinstance(value, (CreateEphemeralIndex, WaitEphemeralIndex)):
            return UiPathResumeTriggerName.INDEX_INGESTION
        if isinstance(value, (CreateBatchTransform, WaitBatchTransform)):
            return UiPathResumeTriggerName.BATCH_RAG
        if isinstance(value, (DocumentExtraction, WaitDocumentExtraction)):
            return UiPathResumeTriggerName.EXTRACTION
        # default to API trigger
        return UiPathResumeTriggerName.API

    async def _handle_task_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle task-type resume triggers.

        Args:
            value: The suspend value (CreateTask or WaitTask)
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.folder_path = value.app_folder_path
        resume_trigger.folder_key = value.app_folder_key

        if isinstance(value, (WaitTask, WaitEscalation)):
            resume_trigger.item_key = value.action.key
        elif isinstance(value, (CreateTask, CreateEscalation)):
            uipath = UiPath()
            action = await uipath.tasks.create_async(
                title=value.title,
                app_name=value.app_name if value.app_name else "",
                app_folder_path=value.app_folder_path if value.app_folder_path else "",
                app_folder_key=value.app_folder_key if value.app_folder_key else "",
                app_key=value.app_key if value.app_key else "",
                assignee=value.assignee if value.assignee else "",
                recipient=value.recipient if value.recipient else "",
                data=value.data,
                priority=value.priority,
                labels=value.labels,
                is_actionable_message_enabled=value.is_actionable_message_enabled,
                actionable_message_metadata=value.actionable_message_metadata,
                source_name=value.source_name,
            )
            if not action:
                raise Exception("Failed to create action")
            resume_trigger.item_key = action.key

    async def _handle_deep_rag_job_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle Deep RAG resume triggers.

        Args:
            value: The suspend value (CreateDeepRag or WaitDeepRag)
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.folder_path = value.index_folder_path
        resume_trigger.folder_key = value.index_folder_key
        if isinstance(value, WaitDeepRag):
            resume_trigger.item_key = value.deep_rag.id
        elif isinstance(value, CreateDeepRag):
            uipath = UiPath()
            if value.is_ephemeral_index:
                deep_rag = (
                    await uipath.context_grounding.start_deep_rag_ephemeral_async(
                        name=value.name,
                        index_id=value.index_id,
                        prompt=value.prompt,
                        glob_pattern=value.glob_pattern,
                        citation_mode=value.citation_mode,
                    )
                )

            else:
                deep_rag = await uipath.context_grounding.start_deep_rag_async(
                    name=value.name,
                    index_name=value.index_name,
                    index_id=value.index_id,
                    prompt=value.prompt,
                    glob_pattern=value.glob_pattern,
                    citation_mode=value.citation_mode,
                    folder_path=value.index_folder_path,
                    folder_key=value.index_folder_key,
                )
            if not deep_rag:
                raise Exception("Failed to start deep rag")

            resume_trigger.item_key = deep_rag.id

        assert resume_trigger.item_key
        await self._create_external_trigger(
            ExternalTrigger(
                type=ExternalTriggerType.DEEP_RAG, external_id=resume_trigger.item_key
            )
        )

    async def _handle_ephemeral_index_job_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle ephemeral index.

        Args:
            value: The suspend value (CreateEphemeralIndex or WaitEphemeralIndex)
            resume_trigger: The resume trigger to populate

        """
        if isinstance(value, WaitEphemeralIndex):
            resume_trigger.item_key = value.index.id
        elif isinstance(value, CreateEphemeralIndex):
            uipath = UiPath()
            ephemeral_index = (
                await uipath.context_grounding.create_ephemeral_index_async(
                    usage=value.usage,
                    attachments=value.attachments,
                )
            )
            if not ephemeral_index:
                raise Exception("Failed to create ephemeral index")
            resume_trigger.item_key = ephemeral_index.id

        assert resume_trigger.item_key
        await self._create_external_trigger(
            ExternalTrigger(
                type=ExternalTriggerType.INDEX_INGESTION,
                external_id=resume_trigger.item_key,
            )
        )

    async def _handle_batch_rag_job_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle batch transform resume triggers.

        Args:
            value: The suspend value (CreateBatchTransform or WaitBatchTransform)
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.folder_path = value.index_folder_path
        resume_trigger.folder_key = value.index_folder_key
        if isinstance(value, WaitBatchTransform):
            resume_trigger.item_key = value.batch_transform.id
        elif isinstance(value, CreateBatchTransform):
            uipath = UiPath()
            if value.is_ephemeral_index:
                batch_transform = await uipath.context_grounding.start_batch_transform_ephemeral_async(
                    name=value.name,
                    index_id=value.index_id,
                    prompt=value.prompt,
                    output_columns=value.output_columns,
                    storage_bucket_folder_path_prefix=value.storage_bucket_folder_path_prefix,
                    enable_web_search_grounding=value.enable_web_search_grounding,
                )
            else:
                batch_transform = await uipath.context_grounding.start_batch_transform_async(
                    name=value.name,
                    index_name=value.index_name,
                    index_id=value.index_id,
                    prompt=value.prompt,
                    output_columns=value.output_columns,
                    storage_bucket_folder_path_prefix=value.storage_bucket_folder_path_prefix,
                    enable_web_search_grounding=value.enable_web_search_grounding,
                    folder_path=value.index_folder_path,
                    folder_key=value.index_folder_key,
                )
            if not batch_transform:
                raise Exception("Failed to start batch transform")

            resume_trigger.item_key = batch_transform.id

        assert resume_trigger.item_key
        await self._create_external_trigger(
            ExternalTrigger(
                type=ExternalTriggerType.BATCH_RAG,
                external_id=resume_trigger.item_key,
            )
        )

    async def _handle_ixp_extraction_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle IXP Extraction resume triggers.

        Args:
            value: The suspend value (DocumentExtraction or WaitDocumentExtraction)
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.folder_path = resume_trigger.folder_key = None
        if isinstance(value, WaitDocumentExtraction):
            resume_trigger.item_key = value.extraction.operation_id
            # add project_id and tag to the payload dict (needed when reading the trigger)
            assert isinstance(resume_trigger.payload, dict)
            resume_trigger.payload.setdefault("project_id", value.extraction.project_id)
            resume_trigger.payload.setdefault("tag", value.extraction.tag)
        elif isinstance(value, DocumentExtraction):
            uipath = UiPath()
            document_extraction = await uipath.documents.start_ixp_extraction_async(
                project_name=value.project_name,
                tag=value.tag,
                file=value.file,
                file_path=value.file_path,
            )
            if not document_extraction:
                raise Exception("Failed to start document extraction")

            resume_trigger.item_key = document_extraction.operation_id

            # add project_id and tag to the payload dict (needed when reading the trigger)
            assert isinstance(resume_trigger.payload, dict)
            resume_trigger.payload.setdefault(
                "project_id", document_extraction.project_id
            )
            resume_trigger.payload.setdefault("tag", document_extraction.tag)

        assert resume_trigger.item_key
        await self._create_external_trigger(
            ExternalTrigger(
                type=ExternalTriggerType.IXP_EXTRACTION,
                external_id=resume_trigger.item_key,
            )
        )

    async def _handle_ixp_vs_escalation_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle IXP VS Escalation resume triggers.

        Args:
            value: The suspend value (DocumentExtractionValidation or WaitDocumentExtractionValidation)
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.folder_path = resume_trigger.folder_key = None

        if isinstance(value, WaitDocumentExtractionValidation):
            resume_trigger.item_key = value.extraction_validation.operation_id

            # add project_id and tag to the payload dict (needed when reading the trigger)
            assert isinstance(resume_trigger.payload, dict)
            resume_trigger.payload.setdefault(
                "project_id", value.extraction_validation.project_id
            )
            resume_trigger.payload.setdefault("tag", value.extraction_validation.tag)
        elif isinstance(value, DocumentExtractionValidation):
            uipath = UiPath()
            extraction_validation = (
                await uipath.documents.start_ixp_extraction_validation_async(
                    extraction_response=value.extraction_response,
                    action_title=value.action_title,
                    action_priority=value.action_priority,
                    action_folder=value.action_folder,
                    storage_bucket_name=value.storage_bucket_name,
                    storage_bucket_directory_path=value.storage_bucket_directory_path,
                )
            )
            if not extraction_validation:
                raise Exception("Failed to start extraction validation")

            resume_trigger.item_key = extraction_validation.operation_id

            # add project_id and tag to the payload dict (needed when reading the trigger)
            assert isinstance(resume_trigger.payload, dict)
            resume_trigger.payload.setdefault(
                "project_id", extraction_validation.project_id
            )
            resume_trigger.payload.setdefault("tag", extraction_validation.tag)

        assert resume_trigger.item_key
        await self._create_external_trigger(
            ExternalTrigger(
                type=ExternalTriggerType.IXP_VS_ESCALATION,
                external_id=resume_trigger.item_key,
            )
        )

    async def _handle_job_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle job-type resume triggers.

        Args:
            value: The suspend value (InvokeProcess, WaitJob, InvokeSystemAgent, WaitSystemAgent)
            resume_trigger: The resume trigger to populate
        """
        if isinstance(value, InvokeSystemAgent):
            resume_trigger.folder_path = value.folder_path
            resume_trigger.folder_key = value.folder_key
        else:
            resume_trigger.folder_path = value.process_folder_path
            resume_trigger.folder_key = value.process_folder_key

        if isinstance(value, WaitJob):
            resume_trigger.item_key = value.job.key
        elif isinstance(value, WaitSystemAgent):
            resume_trigger.item_key = value.job_key
        elif isinstance(value, InvokeProcess):
            uipath = UiPath()
            job = await uipath.processes.invoke_async(
                name=value.name,
                input_arguments=value.input_arguments,
                attachments=value.attachments,
                folder_path=value.process_folder_path,
                folder_key=value.process_folder_key,
            )
            if not job:
                raise Exception("Failed to invoke process")
            resume_trigger.item_key = job.key
        elif isinstance(value, InvokeSystemAgent):
            uipath = UiPath()
            job_key = await uipath.agenthub.invoke_system_agent_async(
                agent_name=value.agent_name,
                entrypoint=value.entrypoint,
                input_arguments=value.input_arguments,
                folder_path=value.folder_path,
                folder_key=value.folder_key,
            )
            if not job_key:
                raise Exception("Failed to invoke system agent")
            resume_trigger.item_key = job_key

    def _handle_api_trigger(
        self, value: Any, resume_trigger: UiPathResumeTrigger
    ) -> None:
        """Handle API-type resume triggers.

        Args:
            value: The suspend value
            resume_trigger: The resume trigger to populate
        """
        resume_trigger.api_resume = UiPathApiTrigger(
            inbox_id=str(uuid.uuid4()), request=serialize_object(value)
        )


class UiPathResumeTriggerHandler:
    """Combined handler for creating and reading resume triggers.

    Implements UiPathResumeTriggerProtocol by composing the creator and reader.
    """

    def __init__(self):
        """Initialize the handler with creator and reader instances."""
        self._creator = UiPathResumeTriggerCreator()
        self._reader = UiPathResumeTriggerReader()

    async def create_trigger(self, suspend_value: Any) -> UiPathResumeTrigger:
        """Create a resume trigger from a suspend value.

        Args:
            suspend_value: The value that caused the suspension.

        Returns:
            UiPathResumeTrigger ready to be persisted

        Raises:
            UiPathRuntimeError: If trigger creation fails
        """
        return await self._creator.create_trigger(suspend_value)

    async def read_trigger(self, trigger: UiPathResumeTrigger) -> Any | None:
        """Read a resume trigger and convert it to runtime-compatible input.

        Args:
            trigger: The resume trigger to read

        Returns:
            The data retrieved from UiPath services, or None if no data is available.

        Raises:
            UiPathRuntimeError: If reading fails or job failed
        """
        return await self._reader.read_trigger(trigger)



class SecretMissingError(Exception):
    """Raised when access token is not configured.

    This exception is raised when attempting to use the SDK without setting
    the access token via the UIPATH_ACCESS_TOKEN environment variable or through authentication.
    """

    def __init__(
        self,
        message="Authentication required. Please run \033[1muipath auth\033[22m or set the UIPATH_ACCESS_TOKEN environment variable to a valid access token.",
    ):
        self.message = message
        super().__init__(self.message)



"""UiPath LLM Gateway Services.

This module provides services for interacting with UiPath's LLM (Large Language Model) Gateway,
offering both OpenAI-compatible and normalized API interfaces for chat completions and embeddings.

The module includes:
- UiPathOpenAIService: OpenAI-compatible API for chat completions and embeddings
- UiPathLlmChatService: UiPath's normalized API with advanced features like tool calling
- ChatModels: Constants for available chat models
- EmbeddingModels: Constants for available embedding models

Classes:
    ChatModels: Container for supported chat model identifiers
    EmbeddingModels: Container for supported embedding model identifiers
    UiPathOpenAIService: Service using OpenAI-compatible API format
    UiPathLlmChatService: Service using UiPath's normalized API format
"""

from typing import Any, Optional

from opentelemetry import trace
from pydantic import BaseModel
from uipath.core.tracing import traced

from ..common._base_service import BaseService
from ..common._config import UiPathApiConfig
from ..common._endpoints_manager import EndpointManager
from ..common._execution_context import UiPathExecutionContext
from ..common._models import Endpoint
from .llm_gateway import (
    ChatCompletion,
    SpecificToolChoice,
    TextEmbedding,
    ToolChoice,
    ToolDefinition,
)
from .llm_throttle import get_llm_semaphore

# Common constants
API_VERSION = "2024-10-21"  # Standard API version for OpenAI-compatible endpoints
NORMALIZED_API_VERSION = (
    "2024-08-01-preview"  # API version for UiPath's normalized endpoints
)

DEFAULT_REQUESTING_PRODUCT = "uipath-python-sdk"
DEFAULT_REQUESTING_FEATURE = "llm-call"


def _build_llm_headers(
    requesting_product: str = DEFAULT_REQUESTING_PRODUCT,
    requesting_feature: str = DEFAULT_REQUESTING_FEATURE,
    agenthub_config: Optional[str] = None,
    action_id: Optional[str] = None,
) -> dict[str, str]:
    headers: dict[str, str] = {
        "X-UIPATH-STREAMING-ENABLED": "false",
        "X-UiPath-LlmGateway-RequestingProduct": requesting_product,
        "X-UiPath-LlmGateway-RequestingFeature": requesting_feature,
    }
    if agenthub_config:
        headers["X-UiPath-AgentHub-Config"] = agenthub_config
    if action_id:
        headers["X-UiPath-LlmGateway-ActionId"] = action_id
    return headers


class ChatModels(object):
    """Available chat models for LLM Gateway services.

    This class provides constants for the supported chat models that can be used
    with both UiPathOpenAIService and UiPathLlmChatService.
    """

    gpt_4 = "gpt-4"
    gpt_4_1106_Preview = "gpt-4-1106-Preview"
    gpt_4_32k = "gpt-4-32k"
    gpt_4_turbo_2024_04_09 = "gpt-4-turbo-2024-04-09"
    gpt_4_vision_preview = "gpt-4-vision-preview"
    gpt_4o_2024_05_13 = "gpt-4o-2024-05-13"
    gpt_4o_2024_08_06 = "gpt-4o-2024-08-06"
    gpt_4o_mini_2024_07_18 = "gpt-4o-mini-2024-07-18"
    gpt_4_1_mini_2025_04_14 = "gpt-4.1-mini-2025-04-14"
    o3_mini = "o3-mini-2025-01-31"


class EmbeddingModels(object):
    """Available embedding models for LLM Gateway services.

    This class provides constants for the supported embedding models that can be used
    with the embeddings functionality.
    """

    text_embedding_3_large = "text-embedding-3-large"
    text_embedding_ada_002 = "text-embedding-ada-002"


def _cleanup_schema(schema: dict[str, Any]) -> dict[str, Any]:
    """Clean up a JSON schema for use with LLM Gateway.

    This function converts a JSON schema to a format that's
    compatible with the LLM Gateway's JSON schema requirements by removing
    titles and other metadata that might cause validation issues.

    Args:
        schema (dict[str, Any]): an input JSON schema.

    Returns:
        dict: A cleaned JSON schema dictionary suitable for LLM Gateway response_format.

    Examples:
        ```python
        from pydantic import BaseModel

        class Country(BaseModel):
            name: str
            capital: str
            languages: list[str]

        schema = _cleanup_schema(Country.model_json_schema())
        # Returns a clean schema without titles and unnecessary metadata
        ```
    """

    def clean_type(type_def):
        """Clean property definitions by removing titles and cleaning nested items. Additionally, `additionalProperties` is ensured on all objects."""
        cleaned_type = {}
        for key, value in type_def.items():
            if key == "title" or key == "properties":
                continue
            else:
                cleaned_type[key] = value
        if type_def.get("type") == "object" and "additionalProperties" not in type_def:
            cleaned_type["additionalProperties"] = False

        if "properties" in type_def:
            properties = type_def.get("properties", {})
            for key, value in properties.items():
                properties[key] = clean_type(value)
            cleaned_type["properties"] = properties

        if type_def.get("type") == "object":
            cleaned_type["required"] = list(cleaned_type.get("properties", {}).keys())

        if "$defs" in type_def:
            cleaned_defs = {}
            for key, value in type_def["$defs"].items():
                cleaned_defs[key] = clean_type(value)
            cleaned_type["$defs"] = cleaned_defs
        return cleaned_type

    # Create clean schema
    clean_schema = clean_type(schema)
    return clean_schema


class UiPathOpenAIService(BaseService):
    """Service for calling UiPath's LLM Gateway using OpenAI-compatible API.

    This service provides access to Large Language Model capabilities through UiPath's
    LLM Gateway, including chat completions and text embeddings. It uses the OpenAI-compatible
    API format and is suitable for applications that need direct OpenAI API compatibility.
    """

    def __init__(
        self,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        *,
        requesting_product: str = DEFAULT_REQUESTING_PRODUCT,
        requesting_feature: str = DEFAULT_REQUESTING_FEATURE,
        agenthub_config: Optional[str] = None,
        action_id: Optional[str] = None,
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)
        self._llm_headers = _build_llm_headers(
            requesting_product, requesting_feature, agenthub_config, action_id
        )

    @traced(name="llm_embeddings", run_type="uipath")
    async def embeddings(
        self,
        input: str,
        embedding_model: str = EmbeddingModels.text_embedding_ada_002,
        openai_api_version: str = API_VERSION,
    ):
        """Generate text embeddings using UiPath's LLM Gateway service.

        This method converts input text into dense vector representations that can be used
        for semantic search, similarity calculations, and other NLP tasks.

        Args:
            input (str): The input text to embed. Can be a single sentence, paragraph,
                or document that you want to convert to embeddings.
            embedding_model (str, optional): The embedding model to use.
                Defaults to EmbeddingModels.text_embedding_ada_002.
                Available models are defined in the EmbeddingModels class.
            openai_api_version (str, optional): The OpenAI API version to use.
                Defaults to API_VERSION.

        Returns:
            TextEmbedding: The embedding response containing the vector representation
                of the input text along with metadata.

        Examples:
            ```python
            # Basic embedding
            embedding = await service.embeddings("Hello, world!")

            # Using a specific model
            embedding = await service.embeddings(
                "This is a longer text to embed",
                embedding_model=EmbeddingModels.text_embedding_3_large
            )
            ```
        """
        endpoint = EndpointManager.get_embeddings_endpoint().format(
            model=embedding_model, api_version=openai_api_version
        )
        endpoint = Endpoint("/" + endpoint)

        async with get_llm_semaphore():
            response = await self.request_async(
                "POST",
                endpoint,
                json={"input": input},
                params={"api-version": API_VERSION},
                headers=self._llm_headers,
            )

        return TextEmbedding.model_validate(response.json())

    @traced(name="LLM call", run_type="uipath")
    async def chat_completions(
        self,
        messages: list[dict[str, str]],
        model: str = ChatModels.gpt_4_1_mini_2025_04_14,
        max_tokens: int = 4096,
        temperature: float = 0,
        response_format: dict[str, Any] | type[BaseModel] | None = None,
        api_version: str = API_VERSION,
    ):
        """Generate chat completions using UiPath's LLM Gateway service.

        This method provides conversational AI capabilities by sending a series of messages
        to a language model and receiving a generated response. It supports multi-turn
        conversations and various OpenAI-compatible models.

        Args:
            messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content' keys.
                The supported roles are 'system', 'user', and 'assistant'. System messages set
                the behavior/context, user messages are from the human, and assistant messages
                are from the AI.
            model (str, optional): The model to use for chat completion.
                Defaults to ChatModels.gpt_4_1_mini_2025_04_14.
                Available models are defined in the ChatModels class.
            max_tokens (int, optional): Maximum number of tokens to generate in the response.
                Defaults to 4096. Higher values allow longer responses.
            temperature (float, optional): Temperature for sampling, between 0 and 1.
                Lower values (closer to 0) make output more deterministic and focused,
                higher values make it more creative and random. Defaults to 0.
            response_format (Optional[Union[Dict[str, Any], type[BaseModel]]], optional):
                An object specifying the format that the model must output. Can be either:
                - A dictionary with response format configuration (traditional format)
                - A Pydantic BaseModel class (automatically converted to JSON schema)
                Used to enable JSON mode or other structured outputs. Defaults to None.
            api_version (str, optional): The API version to use. Defaults to API_VERSION.

        Returns:
            ChatCompletion: The chat completion response containing the generated message,
                usage statistics, and other metadata.

        Examples:
            ```python
            # Simple conversation
            messages = [
                {"role": "system", "content": "You are a helpful Python programming assistant."},
                {"role": "user", "content": "How do I read a file in Python?"}
            ]
            response = await service.chat_completions(messages)

            # Multi-turn conversation with more tokens
            messages = [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is machine learning?"},
                {"role": "assistant", "content": "Machine learning is a subset of AI..."},
                {"role": "user", "content": "Can you give me a practical example?"}
            ]
            response = await service.chat_completions(
                messages,
                max_tokens=200,
                temperature=0.3
            )

            # Using Pydantic model for structured response
            from pydantic import BaseModel

            class Country(BaseModel):
                name: str
                capital: str
                languages: list[str]

            response = await service.chat_completions(
                messages=[
                    {"role": "system", "content": "You are a helpful assistant. Respond with structured JSON."},
                    {"role": "user", "content": "Tell me about Canada."}
                ],
                response_format=Country,  # Pass BaseModel directly
                max_tokens=1000
            )
            ```

        Note:
            The conversation history can be included to provide context to the model.
            Each message should have both 'role' and 'content' keys.
            When using a Pydantic BaseModel as response_format, it will be automatically
            converted to the appropriate JSON schema format for the LLM Gateway.
        """
        span = trace.get_current_span()
        span.set_attribute("model", model)
        span.set_attribute("uipath.custom_instrumentation", True)

        endpoint = EndpointManager.get_passthrough_endpoint().format(
            model=model, api_version=api_version
        )
        endpoint = Endpoint("/" + endpoint)

        request_body = {
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
        }

        # Handle response_format - convert BaseModel to schema if needed
        if response_format:
            if isinstance(response_format, type) and issubclass(
                response_format, BaseModel
            ):
                # Convert Pydantic model to JSON schema format
                cleaned_schema = _cleanup_schema(response_format.model_json_schema())
                request_body["response_format"] = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": response_format.__name__.lower(),
                        "strict": True,
                        "schema": cleaned_schema,
                    },
                }
            else:
                # Use provided dictionary format directly
                request_body["response_format"] = response_format

        async with get_llm_semaphore():
            response = await self.request_async(
                "POST",
                endpoint,
                json=request_body,
                params={"api-version": API_VERSION},
                headers=self._llm_headers,
            )

        return ChatCompletion.model_validate(response.json())


class UiPathLlmChatService(BaseService):
    """Service for calling UiPath's normalized LLM Gateway API.

    This service provides access to Large Language Model capabilities through UiPath's
    normalized LLM Gateway API. Unlike the OpenAI-compatible service, this service uses
    UiPath's standardized API format and supports advanced features like tool calling,
    function calling, and more sophisticated conversation control.

    The normalized API provides a consistent interface across different underlying model
    providers and includes enhanced features for enterprise use cases.
    """

    def __init__(
        self,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        *,
        requesting_product: str = DEFAULT_REQUESTING_PRODUCT,
        requesting_feature: str = DEFAULT_REQUESTING_FEATURE,
        agenthub_config: Optional[str] = None,
        action_id: Optional[str] = None,
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)
        self._llm_headers = _build_llm_headers(
            requesting_product, requesting_feature, agenthub_config, action_id
        )

    @traced(name="LLM call", run_type="uipath")
    async def chat_completions(
        self,
        messages: list[dict[str, str]] | list[tuple[str, str]],
        model: str = ChatModels.gpt_4_1_mini_2025_04_14,
        max_tokens: int = 4096,
        temperature: float = 0,
        n: int = 1,
        frequency_penalty: float = 0,
        presence_penalty: float = 0,
        top_p: float | None = 1,
        top_k: int | None = None,
        tools: list[ToolDefinition] | None = None,
        tool_choice: ToolChoice | None = None,
        response_format: dict[str, Any] | type[BaseModel] | None = None,
        api_version: str = NORMALIZED_API_VERSION,
    ):
        """Generate chat completions using UiPath's normalized LLM Gateway API.

        This method provides advanced conversational AI capabilities with support for
        tool calling, function calling, and sophisticated conversation control parameters.
        It uses UiPath's normalized API format for consistent behavior across different
        model providers.

        Args:
            messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content' keys.
                The supported roles are 'system', 'user', and 'assistant'. System messages set
                the behavior/context, user messages are from the human, and assistant messages
                are from the AI.
            model (str, optional): The model to use for chat completion.
                Defaults to ChatModels.gpt_4_1_mini_2025_04_14.
                Available models are defined in the ChatModels class.
            max_tokens (int, optional): Maximum number of tokens to generate in the response.
                Defaults to 4096. Higher values allow longer responses.
            temperature (float, optional): Temperature for sampling, between 0 and 1.
                Lower values (closer to 0) make output more deterministic and focused,
                higher values make it more creative and random. Defaults to 0.
            n (int, optional): Number of chat completion choices to generate for each input.
                Defaults to 1. Higher values generate multiple alternative responses.
            frequency_penalty (float, optional): Penalty for token frequency between -2.0 and 2.0.
                Positive values reduce repetition of frequent tokens. Defaults to 0.
            presence_penalty (float, optional): Penalty for token presence between -2.0 and 2.0.
                Positive values encourage discussion of new topics. Defaults to 0.
            top_p (float, optional): Nucleus sampling parameter between 0 and 1.
                Controls diversity by considering only the top p probability mass. Defaults to 1.
            top_k (int, optional): Nucleus sampling parameter.
                Controls diversity by considering only the top k most probable tokens. Defaults to None.
            tools (Optional[List[ToolDefinition]], optional): List of tool definitions that the
                model can call. Tools enable the model to perform actions or retrieve information
                beyond text generation. Defaults to None.
            tool_choice (Optional[ToolChoice], optional): Controls which tools the model can call.
                Can be "auto" (model decides), "none" (no tools), or a specific tool choice.
                Defaults to None.
            response_format (Optional[Union[Dict[str, Any], type[BaseModel]]], optional):
                An object specifying the format that the model must output. Can be either:
                - A dictionary with response format configuration (traditional format)
                - A Pydantic BaseModel class (automatically converted to JSON schema)
                Used to enable JSON mode or other structured outputs. Defaults to None.
            api_version (str, optional): The normalized API version to use.
                Defaults to NORMALIZED_API_VERSION.

        Returns:
            ChatCompletion: The chat completion response containing the generated message(s),
                tool calls (if any), usage statistics, and other metadata.

        Examples:
            ```python
            # Basic conversation
            messages = [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the weather like today?"}
            ]
            response = await service.chat_completions(messages)

            # Conversation with tool calling
            tools = [
                ToolDefinition(
                    function=FunctionDefinition(
                        name="get_weather",
                        description="Get current weather for a location",
                        parameters=ParametersDefinition(
                            type="object",
                            properties={
                                "location": PropertyDefinition(
                                    type="string",
                                    description="City name"
                                )
                            },
                            required=["location"]
                        )
                    )
                )
            ]
            response = await service.chat_completions(
                messages,
                tools=tools,
                tool_choice="auto",
                max_tokens=500
            )

            # Advanced parameters for creative writing
            response = await service.chat_completions(
                messages,
                temperature=0.8,
                top_p=0.9,
                frequency_penalty=0.3,
                presence_penalty=0.2,
                n=3  # Generate 3 alternative responses
            )

            # Using Pydantic model for structured response
            from pydantic import BaseModel

            class Country(BaseModel):
                name: str
                capital: str
                languages: list[str]

            response = await service.chat_completions(
                messages=[
                    {"role": "system", "content": "You are a helpful assistant. Respond with structured JSON."},
                    {"role": "user", "content": "Tell me about Canada."}
                ],
                response_format=Country,  # Pass BaseModel directly
                max_tokens=1000
            )
            )
            ```

        Note:
            This service uses UiPath's normalized API format which provides consistent
            behavior across different underlying model providers and enhanced enterprise features.
        """
        span = trace.get_current_span()
        span.set_attribute("model", model)
        span.set_attribute("uipath.custom_instrumentation", True)

        converted_messages = []

        for message in messages:
            if isinstance(message, tuple) and len(message) == 2:
                role, content = message
                converted_messages.append({"role": role, "content": content})
            elif isinstance(message, dict):
                converted_messages.append(message)
            else:
                raise ValueError(
                    f"Invalid message format: {message}. Expected tuple (role, content) or dict with 'role' and 'content' keys."
                )

        endpoint = EndpointManager.get_normalized_endpoint().format(
            model=model, api_version=api_version
        )
        endpoint = Endpoint("/" + endpoint)

        # Build request body - Claude models don't support some OpenAI-specific parameters
        is_claude_model = "claude" in model.lower()

        request_body = {
            "messages": converted_messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
        }

        # Only add OpenAI-specific parameters for non-Claude models
        if not is_claude_model:
            request_body["n"] = n
            request_body["frequency_penalty"] = frequency_penalty
            request_body["presence_penalty"] = presence_penalty
            if top_p is not None:
                request_body["top_p"] = top_p

        if top_k is not None:
            request_body["top_k"] = top_k

        # Handle response_format - convert BaseModel to schema if needed
        if response_format:
            if isinstance(response_format, type) and issubclass(
                response_format, BaseModel
            ):
                # Convert Pydantic model to JSON schema format
                cleaned_schema = _cleanup_schema(response_format.model_json_schema())
                request_body["response_format"] = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": response_format.__name__.lower(),
                        "strict": True,
                        "schema": cleaned_schema,
                    },
                }
            else:
                # Use provided dictionary format directly
                request_body["response_format"] = response_format

        # Add tools if provided - convert to UiPath format
        if tools:
            request_body["tools"] = [
                self._convert_tool_to_uipath_format(tool) for tool in tools
            ]

        # Handle tool_choice
        if tool_choice:
            if isinstance(tool_choice, str):
                request_body["tool_choice"] = tool_choice
            elif isinstance(tool_choice, SpecificToolChoice):
                request_body["tool_choice"] = {"type": "tool", "name": tool_choice.name}
            else:
                request_body["tool_choice"] = tool_choice.model_dump()

        headers = {
            **self._llm_headers,
            "X-UiPath-LlmGateway-NormalizedApi-ModelName": model,
            "X-UiPath-LLMGateway-AllowFull4xxResponse": "true",
        }

        # Log the complete request for debugging
        import json as json_module
        import logging

        logger = logging.getLogger(__name__)

        logger.info("=" * 80)
        logger.info("📤 LLM Gateway Normalized API Request")
        logger.info("=" * 80)
        logger.info(f"Model: {model}")
        logger.info(f"Endpoint: {endpoint}")
        logger.info(f"API Version: {NORMALIZED_API_VERSION}")
        logger.info(f"Is Claude Model: {is_claude_model}")
        logger.info("-" * 80)
        logger.info("Headers:")
        for key, value in headers.items():
            logger.info(f"  {key}: {value}")
        logger.info("-" * 80)
        logger.info("Request Body:")
        # Create a copy for logging with tools truncated for readability
        log_body: dict[str, Any] = request_body.copy()
        tools_list = log_body.get("tools")
        if tools_list and isinstance(tools_list, list):
            log_body["tools"] = f"[{len(tools_list)} tool(s)]"
        messages_list = log_body.get("messages")
        if messages_list and isinstance(messages_list, list):
            log_body["messages"] = [
                {
                    **msg,
                    "content": msg["content"][:100] + "..."
                    if len(msg.get("content", "")) > 100
                    else msg["content"],
                }
                for msg in messages_list
            ]
        logger.info(json_module.dumps(log_body, indent=2))
        logger.info("=" * 80)

        async with get_llm_semaphore():
            response = await self.request_async(
                "POST",
                endpoint,
                json=request_body,
                params={"api-version": NORMALIZED_API_VERSION},
                headers=headers,
            )

        logger.info(f"✅ Response received with status: {response.status_code}")
        return ChatCompletion.model_validate(response.json())

    def _convert_tool_to_uipath_format(self, tool: ToolDefinition) -> dict[str, Any]:
        """Convert an OpenAI-style tool definition to UiPath API format.

        This internal method transforms tool definitions from the standard OpenAI format
        to the format expected by UiPath's normalized LLM Gateway API.

        Args:
            tool (ToolDefinition): The tool definition in OpenAI format containing
                function name, description, and parameter schema.

        Returns:
            Dict[str, Any]: The tool definition converted to UiPath API format
                with the appropriate structure and field mappings.
        """
        parameters = {
            "type": tool.function.parameters.type,
            "properties": {
                name: {
                    "type": prop.type,
                    **({"description": prop.description} if prop.description else {}),
                    **({"enum": prop.enum} if prop.enum else {}),
                }
                for name, prop in tool.function.parameters.properties.items()
            },
        }

        if tool.function.parameters.required:
            parameters["required"] = tool.function.parameters.required

        return {
            "name": tool.function.name,
            "description": tool.function.description,
            "parameters": parameters,
        }



"""UiPath Common Models.

This module contains common models used across multiple services.
"""

from ._api_client import ApiClient
from ._base_service import BaseService
from ._bindings import (
    ConnectionResourceOverwrite,
    GenericResourceOverwrite,
    ResourceOverwrite,
    ResourceOverwriteParser,
    ResourceOverwritesContext,
    resource_override,
)
from ._config import UiPathApiConfig, UiPathConfig
from ._endpoints_manager import EndpointManager
from ._execution_context import UiPathExecutionContext
from ._external_application_service import ExternalApplicationService
from ._folder_context import FolderContext, header_folder
from ._http_config import get_httpx_client_kwargs
from ._models import Endpoint, RequestSpec
from ._service_url_overrides import inject_routing_headers, resolve_service_url
from ._span_utils import UiPathSpan, _SpanUtils
from ._url import UiPathUrl
from ._user_agent import user_agent_value
from .auth import TokenData
from .dynamic_schema import jsonschema_to_pydantic
from .interrupt_models import (
    CreateBatchTransform,
    CreateDeepRag,
    CreateDeepRagRaw,
    CreateEphemeralIndex,
    CreateEphemeralIndexRaw,
    CreateEscalation,
    CreateTask,
    DocumentExtraction,
    DocumentExtractionValidation,
    InvokeProcess,
    InvokeProcessRaw,
    InvokeSystemAgent,
    WaitBatchTransform,
    WaitDeepRag,
    WaitDeepRagRaw,
    WaitDocumentExtraction,
    WaitDocumentExtractionValidation,
    WaitEphemeralIndex,
    WaitEphemeralIndexRaw,
    WaitEscalation,
    WaitJob,
    WaitJobRaw,
    WaitSystemAgent,
    WaitTask,
)
from .paging import PagedResult

__all__ = [
    "ApiClient",
    "BaseService",
    "UiPathApiConfig",
    "UiPathExecutionContext",
    "ExternalApplicationService",
    "FolderContext",
    "TokenData",
    "UiPathConfig",
    "CreateTask",
    "CreateEscalation",
    "WaitEscalation",
    "InvokeProcess",
    "InvokeProcessRaw",
    "WaitTask",
    "WaitJob",
    "WaitJobRaw",
    "PagedResult",
    "CreateDeepRag",
    "CreateDeepRagRaw",
    "WaitDeepRag",
    "WaitDeepRagRaw",
    "CreateBatchTransform",
    "WaitBatchTransform",
    "DocumentExtraction",
    "WaitDocumentExtraction",
    "InvokeSystemAgent",
    "WaitSystemAgent",
    "CreateEphemeralIndex",
    "CreateEphemeralIndexRaw",
    "WaitEphemeralIndex",
    "WaitEphemeralIndexRaw",
    "DocumentExtractionValidation",
    "WaitDocumentExtractionValidation",
    "RequestSpec",
    "Endpoint",
    "UiPathUrl",
    "user_agent_value",
    "get_httpx_client_kwargs",
    "resource_override",
    "header_folder",
    "validate_pagination_params",
    "EndpointManager",
    "jsonschema_to_pydantic",
    "ConnectionResourceOverwrite",
    "GenericResourceOverwrite",
    "ResourceOverwrite",
    "ResourceOverwriteParser",
    "ResourceOverwritesContext",
    "UiPathSpan",
    "_SpanUtils",
    "resolve_service_url",
    "inject_routing_headers",
]

from .validation import validate_pagination_params
